@@ -8,7 +8,8 @@
 #include <sys/select.h>
 
 void cap_init(const char *device, const char *filter);
-void cap_fd_set(fd_set *read_set, int *max_fd);
+void cap_fd_set(fd_set *read_set, int *max_fd,
+   struct timeval *timeout, int *need_timeout);
 void cap_poll(fd_set *read_set);
 void cap_stop(void);
 
