@@ -10,6 +10,7 @@
 #include "darkstat.h"
 #include "acct.h"
 #include "cap.h"
+#include "conv.h"
 #include "dns.h"
 #include "graph_db.h"
 #include "http.h"
@@ -53,31 +54,11 @@ parse_port(const char *str)
 static void
 usage(const char *arg0)
 {
-   printf("usage: %s -i interface [-p port] [-b bindaddr] [-f filter]\n",
+   printf("usage: %s [-d] -i interface [-p port] [-b bindaddr] [-f filter]\n",
       arg0);
    exit(EXIT_FAILURE);
 }
 
-static void
-reduce_poll(void)
-{
-   static time_t last = 0;
-   /* FIXME: make more sane/tunable */
-   #define REDUCE_PERIOD 30 /* every <n> seconds */
-   #define REDUCE_HOSTS 100 /* reduce DB to <n> hosts */
-
-   if (last == 0) {
-      last = time(NULL);
-      return;
-   }
-
-   /* FIXME: use monotonic clock for this */
-   if (time(NULL) - last > REDUCE_PERIOD) {
-      hosts_db_reduce(REDUCE_HOSTS);
-      last = time(NULL);
-   }
-}
-
 int
 main(int argc, char **argv)
 {
@@ -87,7 +68,7 @@ main(int argc, char **argv)
    int maxconn = -1;
    const char *interface = NULL;
    const char *filter = NULL;
-   int ch;
+   int ch, want_daemonize = 0;
 
    printf(PACKAGE_STRING " using libpcap %d.%d\n",
       PCAP_VERSION_MAJOR, PCAP_VERSION_MINOR);
@@ -95,8 +76,11 @@ main(int argc, char **argv)
    if (close(STDIN_FILENO) == -1)
       warn("close(stdin)");
 
-   while ((ch = getopt(argc, argv, "i:p:b:f:")) != -1) {
+   while ((ch = getopt(argc, argv, "di:p:b:f:")) != -1) {
    switch (ch) {
+   case 'd':
+      want_daemonize = 1;
+      break;
    case 'i':
       interface = optarg;
       break;
@@ -123,12 +107,21 @@ main(int argc, char **argv)
       usage(argv[0]);
    }
 
+   if (want_daemonize)
+      daemonize_start();
+
    dns_init(); /* do this first as it forks - minimize memory use */
-   ncache_init();
+
+   /* Need root privs for these: */
+   cap_init(interface, filter);
+   http_init(bindaddr, bindport, maxconn);
+   ncache_init(); /* don't need priv, but must do before chroot */
+
+   privdrop(1 /* and chroot */);
+
+   /* Don't need root privs for these: */
    graph_init();
    hosts_db_init();
-   http_init(bindaddr, bindport, maxconn);
-   cap_init(interface, filter);
    localip_init(interface);
 
    if (signal(SIGTERM, sig_shutdown) == SIG_ERR)
@@ -136,34 +129,34 @@ main(int argc, char **argv)
    if (signal(SIGINT, sig_shutdown) == SIG_ERR)
       errx(1, "signal() failed");
 
-   verbosef("main loop");
+   verbosef("entering main loop");
+   daemonize_finish();
 
    while (running) {
-      int select_ret, max_fd = -1, use_timeout;
+      int select_ret, max_fd = -1, use_timeout = 0;
       struct timeval timeout;
       fd_set rs, ws;
 
       FD_ZERO(&rs);
       FD_ZERO(&ws);
 
-      cap_fd_set(&rs, &max_fd);
+      cap_fd_set(&rs, &max_fd, &timeout, &use_timeout);
       http_fd_set(&rs, &ws, &max_fd, &timeout, &use_timeout);
 
       select_ret = select(max_fd+1, &rs, &ws, NULL,
          (use_timeout) ? &timeout : NULL);
 
-      if (select_ret == 0) {
-         if (!use_timeout)
-            errx(1, "select() timed out");
-      }
-      else if (select_ret == -1) {
+      if ((select_ret == 0) && (!use_timeout))
+            errx(1, "select() erroneously timed out");
+
+      if (select_ret == -1) {
          if (errno == EINTR)
             continue;
          else
             err(1, "select()");
       }
       else {
-         reduce_poll();
+         hosts_db_reduce();
          graph_rotate();
          cap_poll(&rs);
          dns_poll();
@@ -177,6 +170,7 @@ main(int argc, char **argv)
    hosts_db_free();
    graph_free();
    ncache_free();
+   verbosef("shut down");
    return (EXIT_SUCCESS);
 }
 
