@@ -18,6 +18,8 @@
 
 #include <sys/socket.h>
 #include <sys/wait.h>
+#include <assert.h>
+#include <errno.h>
 #include <netdb.h>
 #include <signal.h>
 #include <stdlib.h>
@@ -31,6 +33,16 @@ static void dns_main(void); /* this is what the child process runs */
 static int sock[2];
 static pid_t pid = -1;
 
+#define DNS_REPLY_LEN 1024
+union dns_reply {
+   char buf[DNS_REPLY_LEN];
+   struct {
+      in_addr_t ip;
+      int error;              /* h_errno, or 0 if no error */
+      char name[];            /* the rest of the buffer is the name */
+   } hdr;
+};
+
 void
 dns_init(void)
 {
@@ -43,14 +55,19 @@ dns_init(void)
 
    if (pid == 0) {
       /* We are the child. */
+      privdrop(0 /* don't chroot */);
       close(sock[PARENT]);
+      sock[PARENT] = -1;
+      daemonize_finish(); /* drop our copy of the lifeline! */
       dns_main();
       verbosef("fell out of dns_main()");
       exit(0);
    } else {
       /* We are the parent. */
       close(sock[CHILD]);
-      verbosef("DNS parent has child with PID %d", pid);
+      sock[CHILD] = -1;
+      fd_set_nonblock(sock[PARENT]);
+      verbosef("DNS child has PID %d", pid);
    }
 }
 
@@ -90,6 +107,7 @@ void
 dns_queue(const in_addr_t ip)
 {
    struct tree_rec *rec;
+   ssize_t num_w;
 
    rec = xmalloc(sizeof(*rec));
    rec->ip = ip;
@@ -98,12 +116,16 @@ dns_queue(const in_addr_t ip)
       free(rec);
       return;
    }
-   /* FIXME: will this block? */
-   write(sock[PARENT], &ip, sizeof(ip));
-}
+   /* FIXME: else: we never free this from ip_tree */
 
-static int pending(const int fd);
-static void xread(const int d, void *buf, const size_t nbytes);
+   num_w = write(sock[PARENT], &ip, sizeof(ip)); /* won't block */
+   if (num_w == 0)
+      warnx("dns_queue: write: ignoring end of file");
+   else if (num_w == -1)
+      warn("dns_queue: ignoring write error");
+   else if (num_w != sizeof(ip))
+      err(1, "dns_queue: wrote %d instead of %d", num_w, sizeof(ip));
+}
 
 /*
  * Returns non-zero if result waiting, stores IP and name into given pointers
@@ -112,20 +134,34 @@ static void xread(const int d, void *buf, const size_t nbytes);
 static int
 dns_get_result(in_addr_t *ip, char **name)
 {
-   int len;
+   union dns_reply reply;
+   ssize_t numread;
 
-   if (!pending(sock[PARENT]))
-      return (0);
+   numread = read(sock[PARENT], &reply, sizeof(reply));
+   if (numread == -1) {
+      if (errno == EAGAIN)
+         return (0); /* no input waiting */
+      else
+         goto error;
+   }
+   if (numread == 0)
+      goto error; /* EOF */
+   if (numread != sizeof(reply))
+      errx(1, "dns_get_result read expected %d, got %d",
+         sizeof(reply), numread);
 
-   xread(sock[PARENT], ip, sizeof(*ip));
-   xread(sock[PARENT], &len, sizeof(len));
-   if (len == -1)
-      return (0); /* resolution failed */
-
-   *name = xmalloc(len+1);
-   xread(sock[PARENT], *name, len);
-   (*name)[len] = '\0';
+   /* Return successful reply. */
+   *ip = reply.hdr.ip;
+   if (reply.hdr.error != 0)
+      xasprintf(name, "(%s)", hstrerror(reply.hdr.error));
+   else
+      *name = xstrdup(reply.hdr.name);
    return (1);
+
+error:
+   warn("dns_get_result: ignoring read error");
+   /* FIXME: re-align to stream?  restart dns child? */
+   return (0);
 }
 
 void
@@ -186,19 +222,6 @@ dequeue(in_addr_t *ip)
    return 1;
 }
 
-static void
-xread(const int d, void *buf, const size_t nbytes)
-{
-   ssize_t ret = read(d, buf, nbytes);
-
-   if (ret == 0)
-      errx(1, "read: end of file");
-   if (ret == -1)
-      err(1, "read");
-   if (ret != (ssize_t)nbytes)
-      err(1, "read %d bytes instead of all %d bytes", ret, nbytes);
-}
-
 static void
 xwrite(const int d, const void *buf, const size_t nbytes)
 {
@@ -210,76 +233,77 @@ xwrite(const int d, const void *buf, const size_t nbytes)
       err(1, "wrote %d bytes instead of all %d bytes", ret, nbytes);
 }
 
-/* Returns non-zero if there is data to be read from <fd>. */
-static int
-pending(const int fd)
-{
-   fd_set rd;
-   struct timeval timeout = { 0, 0 };
-
-   FD_ZERO(&rd);
-   FD_SET(fd, &rd);
-   if (select(fd+1, &rd, NULL, NULL, &timeout) == -1)
-      err(1, "select");
-   return (FD_ISSET(fd, &rd));
-}
-
-/* Sleep on an FD until there is data to be read from it. */
-static void
-sleep_on(const int fd)
-{
-   fd_set rd;
-
-   FD_ZERO(&rd);
-   FD_SET(fd, &rd);
-   if (select(fd+1, &rd, NULL, NULL, NULL) == -1)
-      err(1, "select");
-}
-
 static void
 dns_main(void)
 {
    in_addr_t ip;
-   ssize_t numread;
 
 #ifdef HAVE_SETPROCTITLE
    setproctitle("DNS child");
 #endif
-   verbosef("DNS child is alive");
+   fd_set_nonblock(sock[CHILD]);
+   verbosef("DNS child entering main DNS loop");
    for (;;) {
-      if (STAILQ_EMPTY(&queue))
-         sleep_on(sock[CHILD]);
+      int blocking;
 
-      while (pending(sock[CHILD])) {
-         /* FIXME: can this block? */
-         numread = read(sock[CHILD], &ip, sizeof(ip));
+      if (STAILQ_EMPTY(&queue)) {
+         blocking = 1;
+         fd_set_block(sock[CHILD]);
+         verbosef("entering blocking read loop");
+      } else {
+         blocking = 0;
+         fd_set_nonblock(sock[CHILD]);
+         verbosef("non-blocking poll");
+      }
+      for (;;) {
+         /* While we have input to process... */
+         ssize_t numread = read(sock[CHILD], &ip, sizeof(ip));
          if (numread == 0)
             exit(0); /* end of file, nothing more to do here. */
-         if (numread == -1)
+         if (numread == -1) {
+            if (!blocking && (errno == EAGAIN))
+               break; /* ran out of input */
+            /* else error */
             err(1, "DNS: read failed");
+         }
          if (numread != sizeof(ip))
             err(1, "DNS: read expecting %d bytes, got %d", sizeof(ip), numread);
          enqueue(ip);
+         if (blocking) {
+            /* After one blocking read, become non-blocking so that when we
+             * run out of input we fall through to queue processing.
+             */
+            blocking = 0;
+            fd_set_nonblock(sock[CHILD]);
+         }
       }
 
       /* Process queue. */
       if (dequeue(&ip)) {
+         union dns_reply reply;
          struct hostent *he;
-         int len;
 
+         reply.hdr.ip = ip;
          he = gethostbyaddr((char *)&ip, sizeof(ip), AF_INET);
-         if (he == NULL) {
-            verbosef("DNS: gethostbyaddr(%s) failed: %s",
-               ip_to_str(ip), hstrerror(h_errno));
-            len = -1;
-         } else
-            len = strlen(he->h_name);
 
-         /* Return result to parent. */
-         xwrite(sock[CHILD], &ip, sizeof(ip));
-         xwrite(sock[CHILD], &len, sizeof(len));
-         if (len != -1)
-            xwrite(sock[CHILD], he->h_name, len);
+         /* On some platforms (for example Linux with GLIBC 2.3.3), h_errno
+          * will be non-zero here even though the lookup succeeded.
+          */
+         if (he == NULL) {
+            reply.hdr.name[0] = '\0';
+            reply.hdr.error = h_errno;
+         } else {
+            char *start = reply.hdr.name;
+            char *end = reply.buf + sizeof(reply);
+            size_t maxlen = end - start;
+
+            strlcpy(reply.hdr.name, he->h_name, maxlen);
+            reply.hdr.error = 0;
+         }
+         fd_set_block(sock[CHILD]);
+         xwrite(sock[CHILD], &reply, sizeof(reply));
+         verbosef("DNS: %s is %s", ip_to_str(ip),
+            (h_errno == 0)?reply.hdr.name:hstrerror(h_errno));
       }
    }
 }
