@@ -22,6 +22,16 @@
 #include <stdlib.h>
 #include <string.h> /* memset(), strcmp() */
 
+/* FIXME: specify somewhere more sane/tunable */
+#define MAX_ENTRIES 30 /* in an HTML table rendered from a hashtable */
+
+/*
+ * Hosts table reduction - when the number of hosts hits NUM_HOSTS_HIGH, we
+ * reduce the table to NUM_HOSTS_LOW hosts.
+ */
+#define NUM_HOSTS_HIGH 1000
+#define NUM_HOSTS_LOW  500
+
 typedef uint32_t (hash_func_t)(const struct hashtable *, const void *);
 typedef const void * (key_func_t)(const struct bucket *);
 typedef int (find_func_t)(const struct bucket *, const void *);
@@ -557,20 +567,17 @@ host_search(const char *ipstr)
 }
 
 /* ---------------------------------------------------------------------------
- * Reduce hosts_db to top <num_hosts> hosts.
+ * Reduce hosts_db if necessary.
  */
 void
-hosts_db_reduce(const unsigned int num_hosts)
+hosts_db_reduce(void)
 {
    uint32_t i, pos, rmd;
    const struct bucket **table;
    uint64_t cutoff;
 
-   if (hosts_db->count < num_hosts) {
-      verbosef("hosts_db_reduce: have only %u hosts, waiting for %u",
-         hosts_db->count, num_hosts);
+   if (hosts_db->count < NUM_HOSTS_HIGH)
       return;
-   }
 
    /* Fill table with pointers to buckets in hashtable. */
    table = xmalloc(sizeof(*table) * hosts_db->count);
@@ -582,8 +589,8 @@ hosts_db_reduce(const unsigned int num_hosts)
       }
    }
    assert(pos == hosts_db->count);
-   qsort_buckets(table, hosts_db->count, 0, num_hosts, TOTAL);
-   cutoff = table[num_hosts - 1]->total;
+   qsort_buckets(table, hosts_db->count, 0, NUM_HOSTS_LOW, TOTAL);
+   cutoff = table[NUM_HOSTS_LOW - 1]->total;
    free(table);
 
    /* Remove all hosts with total < cutoff. */
@@ -608,7 +615,8 @@ hosts_db_reduce(const unsigned int num_hosts)
          b = next;
       }
    }
-   verbosef("hosts_db_reduce: removed %u hosts", rmd);
+   verbosef("hosts_db_reduce: from %u, removed %u hosts",
+      hosts_db->count+rmd, rmd);
    hashtable_rehash(hosts_db, hosts_db->bits); /* needed? */
 }
 
@@ -736,7 +744,6 @@ format_table(struct str *buf, struct hashtable *ht)
 {
    const struct bucket **table;
    uint32_t i, pos, top;
-   #define MAX_ENTRIES 20 /* FIXME: specify somewhere more sane/tunable */
 
    if ((ht == NULL) || (ht->count == 0)) {
       str_append(buf, "<p>The table is empty.</p>\n");
