@@ -46,7 +46,7 @@ struct bucket {
 enum sort_dir { IN, OUT, TOTAL };
 
 void hosts_db_init(void);
-void hosts_db_reduce(const unsigned int num_hosts);
+void hosts_db_reduce(void);
 void hosts_db_free(void);
 
 struct bucket *host_find(const in_addr_t ip); /* can return NULL */
